"""Support for communicating with the Stationary Transceiver Unit (STU)"""

# -- Imports ------------------------------------------------------------------

from __future__ import annotations

from argparse import ArgumentTypeError
from asyncio import sleep
from time import monotonic
from types import TracebackType
from typing import NamedTuple

from netaddr import EUI

from icotronic.cmdline.parse import (
    node_name as check_name,
    sensor_node_number as check_sensor_node_number,
)
from icotronic.can.constants import SENSOR_NODE_NUMBER_SELF_ADDRESSING
from icotronic.can.node.eeprom.node import NodeEEPROM
from icotronic.can.error import ErrorResponseError, NoResponseError
from icotronic.can.node.basic import Node
from icotronic.can.node.id import NodeId
from icotronic.can.node.sensor import SensorNode
from icotronic.can.node.spu import SPU
from icotronic.utility.data import convert_bytes_to_text
from icotronic.test.misc import skip_hardware_tests_ci

pytestmark = skip_hardware_tests_ci()


# -- Classes ------------------------------------------------------------------


class AsyncSensorNodeManager:
    """Context manager for connection to sensor node

    Args:

        stu:
            The STU instance that created the context manager

        identifier:
            The identifier of the sensor node

        sensor_node_class:
            The sensor node class returned by the context manager

    Raises:

        ValueError:
             If you use an invalid name or node number as identifier

    """

    def __init__(
        self,
        stu: STU,
        identifier: int | str | EUI,
        sensor_node_class: type[SensorNode] = SensorNode,
    ) -> None:

        try:
            if isinstance(identifier, int):
                check_sensor_node_number(str(identifier))
            elif isinstance(identifier, str):
                check_name(identifier)
        except ArgumentTypeError as error:
            raise ValueError(error) from error

        self.stu = stu
        self.identifier = identifier
        self.sensor_node_class = sensor_node_class

    async def __aenter__(self) -> SensorNode:
        """Create the connection to the sensor node"""

        def get_sensor_node(
            nodes: list[SensorNodeInfo], identifier: int | str | EUI
        ) -> SensorNodeInfo | None:
            """Get the MAC address of a sensor node"""

            for node in nodes:
                if (  # pylint: disable=too-many-boolean-expressions
                    isinstance(identifier, str)
                    and node.name == identifier
                    or isinstance(identifier, int)
                    and node.sensor_node_number == identifier
                    or isinstance(identifier, EUI)
                    and node.mac_address == identifier
                ):
                    return node

            return None

        await self.stu.activate_bluetooth()

        # We wait for a certain amount of time for the connection to the
        # node to take place
        timeout_in_s = 20
        end_time = monotonic() + timeout_in_s

        sensor_node = None
        sensor_nodes: list[SensorNodeInfo] = []
        while sensor_node is None:
            if monotonic() > end_time:
                sensor_nodes_representation = "\n".join(
                    [repr(node) for node in sensor_nodes]
                )
                node_info = (
                    "Found the following sensor nodes:\n"
                    f"{sensor_nodes_representation}"
                    if len(sensor_nodes) > 0
                    else "No sensor nodes found"
                )

                identifier_description = (
                    "MAC address"
                    if isinstance(self.identifier, EUI)
                    else (
                        "sensor_node_number"
                        if isinstance(self.identifier, int)
                        else "name"
                    )
                )
                raise TimeoutError(
                    "Unable to find sensor node with "
                    f"{identifier_description} “{self.identifier}” in "
                    f"{timeout_in_s} seconds\n\n{node_info}"
                )

            sensor_nodes = await self.stu.get_sensor_nodes()
            sensor_node = get_sensor_node(sensor_nodes, self.identifier)
            if sensor_node is None:
                await sleep(0.1)

        connection_attempt_time = monotonic()
        disconnected = True
        while disconnected:
            await self.stu.connect_with_number(sensor_node.sensor_node_number)
            retry_time_s = 3
            end_time_retry = monotonic() + retry_time_s
            while monotonic() < end_time_retry:
                if monotonic() > end_time:
                    connection_time = monotonic() - connection_attempt_time
                    raise TimeoutError(
                        "Unable to connect to sensor node"
                        f" “{sensor_node}” in"
                        f" {connection_time:.3f} seconds"
                    )

                if await self.stu.is_connected():
                    disconnected = False
                    break

                await sleep(0.1)

        return self.sensor_node_class(self.stu.spu)

    async def __aexit__(
        self,
        exception_type: type[BaseException] | None,
        exception_value: BaseException | None,
        traceback: TracebackType | None,
    ) -> None:
        """Disconnect the sensor node and clean up resources

        Args:

            exception_type:
                The type of the exception in case of an exception

            exception_value:
                The value of the exception in case of an exception

            traceback:
                The traceback in case of an exception

        """

        try:
            await self.stu.deactivate_bluetooth()
        except (NoResponseError, ErrorResponseError):
            pass


class SensorNodeInfo(NamedTuple):
    """Used to store information about a (disconnected) STH"""

    name: str
    """The (Bluetooth advertisement) name of the STH"""

    sensor_node_number: int
    """The node number of the STH"""

    mac_address: EUI
    """The (Bluetooth) MAC address of the STH"""

    rssi: int
    """The RSSI of the STH"""

    def __repr__(self) -> str:
        """Return the string representation of an STH

        Returns:

            A textual representation of the sensor node information

        """

        attributes = ", ".join([
            f"Name: {self.name}",
            f"Number: {self.sensor_node_number}",
            f"MAC Address: {self.mac_address}",
            f"RSSI: {self.rssi}",
        ])
        return f"🤖 {attributes}"

    def __hash__(self):
        """Calculate hash value

        Note:

            This function is required in addition to ``__eq__`` to be
            able to put sensor node information into a set.

        Returns:

            The hash value of the sensor device information

        Examples:

            Update old information about sensor devices with new information

            >>> sensor_node_1 = SensorNodeInfo(
            ...                     name="Test-STH",
            ...                     sensor_node_number=1,
            ...                     mac_address=EUI("08-6B-D7-01-DE-81"),
            ...                     rssi=-58 )
            >>> sensor_node_2 = SensorNodeInfo(
            ...                     name="Test-STH",
            ...                     sensor_node_number=2,
            ...                     mac_address=EUI("08-6B-D7-01-DE-81"),
            ...                     rssi=-80)
            >>> sensor_node_3 = SensorNodeInfo(
            ...                     name="Something",
            ...                     sensor_node_number=3,
            ...                     mac_address=EUI("12-34-56-78-9A-BC"),
            ...                     rssi=-80)

            >>> old = {sensor_node_1, sensor_node_3}
            >>> new = {sensor_node_2, sensor_node_3}
            >>> new | old == {sensor_node_2, sensor_node_3}
            True

        """

        return hash(self.mac_address)

    def __eq__(self, other: object) -> bool:
        """Compare two sensor nodes for equality

        Args:

            other:
                The object this sensor node (information) should be compared
                to

        Returns:

            - ``True`` if the sensor nodes share the same name and MAC address,
              or
            - ``False`` otherwise

        Examples:

            Compare two sensor nodes, which are considered equal

            >>> sensor_node_1 = SensorNodeInfo(
            ...                     name="Test-STH",
            ...                     sensor_node_number=1,
            ...                     mac_address=EUI("08-6B-D7-01-DE-81"),
            ...                     rssi=-58 )
            >>> sensor_node_2 = SensorNodeInfo(
            ...                     name="Test-STH",
            ...                     sensor_node_number=2,
            ...                     mac_address=EUI("08-6B-D7-01-DE-81"),
            ...                     rssi=-80)
            >>> sensor_node_1 == sensor_node_2
            True

            Compare two sensor nodes, which are not considered equal

            >>> sensor_node_3 = SensorNodeInfo(
            ...                     name="Something",
            ...                     sensor_node_number=3,
            ...                     mac_address=EUI("12-34-56-78-9A-BC"),
            ...                     rssi=-80)
            >>> sensor_node_2 == sensor_node_3
            False

        """

        if not isinstance(other, type(self)):
            return NotImplemented

        return self.mac_address == other.mac_address


class STU(Node):
    """Communicate and control a connected STU

    Args:

        spu:
            The SPU object that created this STU instance

    Examples:

        Import required library code

        >>> from asyncio import run
        >>> from icotronic.can.connection import Connection

        Create an STU object

        >>> async def create_stu():
        ...     async with Connection() as stu:
        ...         pass # call some coroutines of `stu` object
        >>> run(create_stu())

    """

    def __init__(self, spu: SPU) -> None:

        super().__init__(spu, NodeEEPROM, NodeId("STU 1"))

    async def activate_bluetooth(self) -> None:
        """Activate Bluetooth on the STU

        Examples:

            Import required library code

            >>> from asyncio import run
            >>> from icotronic.can.connection import Connection

            Activate Bluetooth on the STU

            >>> async def activate():
            ...     async with Connection() as stu:
            ...         await stu.activate_bluetooth()
            >>> run(activate())

        """

        await self.spu.request_bluetooth(
            node=self.id,
            subcommand=1,
            description=f"activate Bluetooth of node “{self.id}”",
            response_data=6 * [0],  # type: ignore[arg-type]
        )

    async def deactivate_bluetooth(self) -> None:
        """Deactivate Bluetooth on the STU

        Examples:

            Import required library code

            >>> from asyncio import run
            >>> from icotronic.can.connection import Connection

            Deactivate Bluetooth on STU 1

            >>> async def deactivate_bluetooth():
            ...     async with Connection() as stu:
            ...         await stu.deactivate_bluetooth()
            >>> run(deactivate_bluetooth())

        """

        await self.spu.request_bluetooth(
            node=self.id,
            subcommand=9,
            description=f"deactivate Bluetooth on “{self.id}”",
            response_data=6 * [0],  # type: ignore[arg-type]
        )

    async def get_available_nodes(self) -> int:
        """Retrieve the number of available sensor nodes

        Returns:

            The number of available sensor nodes

        Examples:

            Import required library code

            >>> from asyncio import run, sleep
            >>> from icotronic.can.connection import Connection

            Get the number of available Bluetooth nodes at STU 1

            >>> async def get_number_bluetooth_nodes():
            ...     async with Connection() as stu:
            ...         await stu.activate_bluetooth()
            ...
            ...         # We assume at least one STH is available
            ...         number_sths = 0
            ...         while number_sths <= 0:
            ...             number_sths = await stu.get_available_nodes()
            ...             await sleep(0.1)
            ...
            ...         return number_sths
            >>> run(get_number_bluetooth_nodes()) >= 0
            1

        """

        answer = await self.spu.request_bluetooth(
            node=self.id,
            subcommand=2,
            description=f"get available Bluetooth nodes of node “{self.id}”",
        )

        available_nodes = int(convert_bytes_to_text(answer.data[2:]))

        return available_nodes

    async def get_name(
        self, sensor_node_number: int = SENSOR_NODE_NUMBER_SELF_ADDRESSING
    ) -> str:
        """Retrieve the name of a sensor node

        Args:

            sensor_node_number:
                The number of the Bluetooth node (0 up to the number of
                available nodes - 1); Use the special node number
                ``SENSOR_NODE_NUMBER_SELF_ADDRESSING`` to retrieve the
                name of the STU itself.

        Note:

            You are probably only interested in the name of the STU itself
            (``SENSOR_NODE_NUMBER_SELF_ADDRESSING``), if you want to know the
            advertisement name of the STU in OTA (Over The Air) update mode
            for flashing a new firmware onto the STU.

        Returns:

            The (Bluetooth broadcast) name of the node

        Examples:

            Import required library code

            >>> from asyncio import run
            >>> from icotronic.can.connection import Connection

            Get Bluetooth advertisement name of node “0” from STU 1

            >>> async def get_bluetooth_node_name():
            ...     async with Connection() as stu:
            ...         await stu.activate_bluetooth()
            ...         # We assume that at least one STH is available
            ...         return await stu.get_name(0)
            >>> name = run(get_bluetooth_node_name())
            >>> isinstance(name, str)
            True
            >>> 0 <= len(name) <= 8
            True

        """

        return await self.spu.get_name(
            node=self.id, sensor_node_number=sensor_node_number
        )

    async def connect_with_number(self, sensor_node_number: int = 0) -> bool:
        """Connect to a Bluetooth node using a node number

        Args:

            sensor_node_number:
                The number of the Bluetooth node (0 up to the number of
                available nodes - 1)

        Returns:

            - True, if

              1. in search mode,
              2. at least single node was found,
              3. no legacy mode,
              4. and scanning mode active

            - False, otherwise

        Examples:

            Import required library code

            >>> from asyncio import run
            >>> from icotronic.can.connection import Connection

            Connect to node “0”

            >>> async def connect_bluetooth_sensor_node_number():
            ...     async with Connection() as stu:
            ...         await stu.activate_bluetooth()
            ...         # We assume that at least one STH is available
            ...         connected = before = await stu.is_connected()
            ...         while not connected:
            ...             connected = await stu.connect_with_number(0)
            ...         await stu.deactivate_bluetooth()
            ...         after = await stu.is_connected()
            ...         # Return status of Bluetooth node connect response
            ...         return before, connected, after
            >>> run(connect_bluetooth_sensor_node_number())
            (False, True, False)

        """

        response = await self.spu.request_bluetooth(
            node=self.id,
            subcommand=7,
            sensor_node_number=sensor_node_number,
            description=f"connect to “{sensor_node_number}” from “{self.id}”",
        )

        return bool(response.data[2])

    async def connect_with_mac_address(self, mac_address: EUI) -> None:
        """Connect to a Bluetooth sensor node using its MAC address

        Args:

            mac_address:
                The MAC address of the sensor node

        Examples:

            Import required library code

            >>> from asyncio import run, sleep
            >>> from icotronic.can.connection import Connection

            Connect to a sensor node via its MAC address

            >>> async def get_bluetooth_mac():
            ...     async with Connection() as stu:
            ...         await stu.activate_bluetooth()
            ...         # Wait for Bluetooth activation to take place
            ...         await sleep(2)
            ...         return await stu.get_mac_address(0)
            >>> mac_address = run(get_bluetooth_mac())
            >>> mac_address != EUI(0)
            True

            >>> async def connect(mac_address):
            ...     async with Connection() as stu:
            ...         await stu.deactivate_bluetooth()
            ...         # We assume that at least one STH is available
            ...         connected = before = await stu.is_connected()
            ...         await stu.activate_bluetooth()
            ...         while not connected:
            ...             await stu.connect_with_mac_address(mac_address)
            ...             await sleep(0.1)
            ...             connected = await stu.is_connected()
            ...         await stu.deactivate_bluetooth()
            ...         after = await stu.is_connected()
            ...         # Return status of Bluetooth node connect response
            ...         return before, connected, after
            >>> run(connect(mac_address))
            (False, True, False)

        """

        mac_address_bytes_reversed = list(reversed(mac_address.packed))
        node = "STU 1"
        # The STU returns reversed MAC address once, probably after the
        # connection was established successfully.
        # Otherwise (before and after) connection took place it returns
        # zeroes all the time. This means the return values is not that
        # useful, e.g. for determining if the STH is connected or not.
        await self.spu.request_bluetooth(
            node=node,
            subcommand=18,
            data=mac_address_bytes_reversed,
            description=f"connect to node “{mac_address}” from “{node}”",
        )

    async def is_connected(self) -> bool:
        """Check if the STU is connected to a Bluetooth node

        Returns:

        - True, if a Bluetooth node is connected to the node
        - False, otherwise

        Examples:

            >>> from asyncio import run, sleep
            >>> from icotronic.can.connection import Connection

            Check connection of node “0” to STU

            >>> async def check_bluetooth_connection():
            ...     async with Connection() as stu:
            ...         await stu.activate_bluetooth()
            ...         await sleep(0.1)
            ...         connected_start = await stu.is_connected()
            ...
            ...         # We assume that at least one STH is available
            ...         await stu.connect_with_number(0)
            ...         # Wait for node connection
            ...         connected_between = False
            ...         while not connected_between:
            ...             connected_between = await stu.is_connected()
            ...             await sleep(0.1)
            ...             await stu.connect_with_number(0)
            ...
            ...         # Deactivate Bluetooth connection
            ...         await stu.deactivate_bluetooth()
            ...         # Wait until node is disconnected
            ...         await sleep(0.1)
            ...         connected_after = await stu.is_connected()
            ...
            ...         return (connected_start, connected_between,
            ...                 connected_after)
            >>> run(check_bluetooth_connection())
            (False, True, False)

        """

        response = await self.spu.request_bluetooth(
            node=self.id,
            subcommand=8,
            response_data=[None, *(5 * [0])],
            description=(
                f"check if “{self.id}” is connected to a Bluetooth node"
            ),
        )

        return bool(response.data[2])

    async def get_rssi(self, sensor_node_number: int):
        """Retrieve the RSSI (Received Signal Strength Indication) of an STH

        Args:

            sensor_node_number:
                The number of the Bluetooth node (0 up to the number of
                available nodes)

        Returns:

            The RSSI of the node

        Examples:

            Import required library code

            >>> from asyncio import run
            >>> from icotronic.can.connection import Connection

            Retrieve the RSSI of a disconnected STH

            >>> async def get_bluetooth_rssi():
            ...     async with Connection() as stu:
            ...         await stu.activate_bluetooth()
            ...         # We assume that at least one STH is available
            ...         # Get the RSSI of node “0”
            ...         return await stu.get_rssi(0)
            >>> rssi = run(get_bluetooth_rssi())
            >>> -80 < rssi < 0
            True

        """

        return await self.spu.get_rssi(
            node=self.id, sensor_node_number=sensor_node_number
        )

    async def get_mac_address(
        self, sensor_node_number: int = SENSOR_NODE_NUMBER_SELF_ADDRESSING
    ) -> EUI:
        """Retrieve the MAC address of the STU or a sensor node

        Note:

            Bluetooth needs to be activated before calling this coroutine,
            otherwise an incorrect MAC address will be returned (for sensor
            nodes).

        Args:

            sensor_node_number:
                The node number of the Bluetooth node (0 up to the number of
                available nodes - 1) or ``0x00``
                (``SENSOR_NODE_NUMBER_SELF_ADDRESSING``) to retrieve the MAC
                address of the STU itself

        Returns:

            The MAC address of the specified sensor node

        Examples:

            Import required library code

            >>> from asyncio import run
            >>> from icotronic.can.connection import Connection

            Retrieve the MAC address of STH 1

            >>> async def get_bluetooth_mac():
            ...     async with Connection() as stu:
            ...         await stu.activate_bluetooth()
            ...         return await stu.get_mac_address(0)
            >>> mac_address = run(get_bluetooth_mac())
            >>> isinstance(mac_address, EUI)
            True
            >>> mac_address != EUI(0)
            True

        """

        return await self.spu.get_mac_address(self.id, sensor_node_number)

    async def get_sensor_nodes(self) -> list[SensorNodeInfo]:
        """Retrieve a list of available sensor nodes

        Returns:

            A list of available nodes including node number, name, MAC address
            and RSSI for each node

        Examples:

            Import required library code

            >>> from asyncio import run, sleep
            >>> from netaddr import EUI
            >>> from icotronic.can.connection import Connection

            Retrieve the list of Bluetooth nodes at STU 1

            >>> async def get_sensor_nodes():
            ...     async with Connection() as stu:
            ...         # We assume that at least one sensor node is available
            ...         nodes = []
            ...         while not nodes:
            ...             nodes = await stu.get_sensor_nodes()
            ...             await sleep(0.1)
            ...
            ...         return nodes
            >>> nodes = run(get_sensor_nodes())
            >>> len(nodes) >= 1
            True
            >>> node = nodes[0]

            >>> node.sensor_node_number
            0

            >>> isinstance(node.name, str)
            True
            >>> 0 <= len(node.name) <= 8
            True

            >>> -80 < node.rssi < 0
            True

            >>> isinstance(node.mac_address, EUI)
            True

        """

        await self.activate_bluetooth()
        available_nodes = await self.get_available_nodes()
        nodes = []
        for node in range(available_nodes):
            mac_address = await self.get_mac_address(node)
            rssi = await self.get_rssi(node)
            name = await self.get_name(node)

            nodes.append(
                SensorNodeInfo(
                    sensor_node_number=node,
                    mac_address=mac_address,
                    name=name,
                    rssi=rssi,
                )
            )

        return nodes

    async def collect_sensor_nodes(self, timeout=5) -> list[SensorNodeInfo]:
        """Collect available sensor nodes

        This coroutine collects sensor nodes until either

        - no new sensor node was found or
        - until the given timeout, if no sensor node was found.

        Args:

            timeout:

                The timeout in seconds until this coroutine returns, if no
                sensor node was found at all

        Returns:

            A list of found sensor nodes

        Examples:

            Import required library code

            >>> from asyncio import run
            >>> from icotronic.can.connection import Connection

            Collect available sensor nodes

            >>> async def collect_sensor_nodes():
            ...     async with Connection() as stu:
            ...         return await stu.collect_sensor_nodes()

            >>> # We assume that at least one sensor node is available
            >>> nodes = run(collect_sensor_nodes())
            >>> len(nodes) >= 1
            True

        """

        timeout = monotonic() + timeout
        sensor_nodes: set[SensorNodeInfo] = set()
        sensor_nodes_before: set[SensorNodeInfo] = set()

        # Wait
        # - until timeout, if there are no devices available or
        # - until no new devices have been found in an iteration of the loop
        while (
            len(sensor_nodes) <= 0
            and monotonic() < timeout
            or sensor_nodes != sensor_nodes_before
        ):
            sensor_nodes_before = set(sensor_nodes)
            sensor_nodes = (
                set(await self.get_sensor_nodes()) | sensor_nodes_before
            )
            await sleep(0.5)

        return list(sensor_nodes)

    def connect_sensor_node(
        self,
        identifier: int | str | EUI,
        sensor_node_class: type[SensorNode] = SensorNode,
    ) -> AsyncSensorNodeManager:
        """Connect to a sensor node (e.g. SHA, SMH or STH)

        Args:

            identifier:
                The

                - MAC address (`EUI`),
                - name (`str`), or
                - node number (`int`)

                of the sensor node we want to connect to

            sensor_node_class:
                Sensor node subclass that should be returned by context manager

        Returns:

            A context manager that returns a sensor node object for the
            connected node

        Raises:

            ValueError:
                 If you use an invalid name or node number as identifier

        Examples:

            Import required library code

            >>> from asyncio import run
            >>> from icotronic.can.connection import Connection

            Connect to the sensor node with node number ``0``

            >>> async def connect_sensor_node():
            ...     async with Connection() as stu:
            ...         async with stu.connect_sensor_node(0):
            ...             connected = await stu.is_connected()
            ...         after = await stu.is_connected()
            ...         return (connected, after)
            >>> run(connect_sensor_node())
            (True, False)

        """

        return AsyncSensorNodeManager(self, identifier, sensor_node_class)


# -- Main ---------------------------------------------------------------------

if __name__ == "__main__":
    from doctest import run_docstring_examples

    run_docstring_examples(
        STU.connect_with_mac_address,
        globals(),
        verbose=True,
    )



"""Support for MytooliT protocol calibration commands"""

# -- Imports ------------------------------------------------------------------

from bidict import bidict

from icotronic.utility.types import check_list

# pylint: disable=too-few-public-methods

# -- Class --------------------------------------------------------------------


class CalibrationMeasurementFormat:
    """This class stores message data for the calibration measurement command

    See also:

    https://mytoolit.github.io/Documentation/#command-calibration-measurement

    This class allows you to specify the message bytes directly as first
    argument. Alternatively you can use the other arguments of the
    initializer to  either

    - overwrite specific parts of the given message bytes or
    - create a calibration format without specifying the message bytes.

    Args:

        data:
            A list containing the (first four) bytes of the calibration
            measurement format

        set:
            Specifies if we want to set or retrieve (get) calibration
            measurement data. If you set this value to false (get), then
            the value of `method` will be ignored

        element:
            Specifies the element that should be measured

            Possible values:

            - Data
            - Temperature
            - Voltage
            - VSS
            - VDD
            - Internal
            - OPV

        method:
            Specifies the calibration method (Activate, Deactivate or
            Measure)

        dimension:
            Specifies the measurement dimension respectively the axis
            (1, 2 or 3)

        reference_voltage:
            The reference voltage in Volt

    """

    methods = ["Reserved", "Activate", "Deactivate", "Measure"]
    elements = bidict({
        "Data": 0,
        "Temperature": 1,
        "Voltage": 32,
        "VSS": 96,
        "VDD": 97,
        "Internal": 98,
        "OPV": 99,
    })

    def __init__(  # pylint: disable=too-many-arguments
        self,
        *data: bytearray | list[int],
        # pylint: disable=redefined-builtin
        set: bool | None = None,
        # pylint: enable=redefined-builtin
        element: str | None = None,
        method: str | None = None,
        dimension: int | None = None,
        reference_voltage: float | None = None,
    ) -> None:

        cls = type(self)

        if data:
            data_bytes = list(data[0])
            check_list(data_bytes, 4)
            self.data = data_bytes[0:4] + [0] * 4
        else:
            self.data = [0, 0, 1, 0] + [0] * 4

        if set is not None:
            method_byte = self.data[0]
            # Set get/set to 0
            method_byte &= 0b01111111
            # Set value
            method_byte |= int(set) << 7
            self.data[0] = method_byte

        if method is not None:
            if method not in cls.methods:
                raise ValueError(f"Unknown method “{method}”")
            method_byte = self.data[0]
            # Set method bits to 0
            method_byte &= 0b10011111
            # Set value
            method_byte |= cls.methods.index(method) << 5
            self.data[0] = method_byte

        if element is not None:
            if element not in cls.elements:
                raise ValueError(f"Unknown element “{element}”")
            self.data[1] = cls.elements[element]

        if dimension is not None:
            if dimension not in {1, 2, 3}:
                raise ValueError(f"Unknown dimension “{dimension}”")
            self.data[2] = dimension

        if reference_voltage is not None:
            self.data[3] = round(reference_voltage * 20) & 0xFF

    def __repr__(self) -> str:
        """
        Retrieve the textual representation of the calibration format

        Returns:

            A string that describes the calibration measurement format

        Examples:

            Get textual representation of calibration measurement formats

            >>> CalibrationMeasurementFormat(set=False, element='Data')
            Get, Data, Dimension: 1, Reference Voltage: 0.0 V

            >>> CalibrationMeasurementFormat(set=True, method='Measure',
            ...     element='Temperature', dimension=2, reference_voltage=3.3)
            Set, Measure, Temperature, Dimension: 2, Reference Voltage: 3.3 V

            >>> CalibrationMeasurementFormat([10, 20, 30, 40])
            Get, Unknown Element, Dimension: 30, Reference Voltage: 2.0 V

        """

        cls = type(self)

        method_byte = self.data[0]

        set_values = method_byte >> 7
        element = cls.elements.inverse.get(self.data[1], "Unknown Element")
        dimension = self.data[2]
        reference_voltage = round(self.data[3] / 20, 1)

        parts = [
            "Set" if set_values else "Get",
            element,
            f"Dimension: {dimension}",
            f"Reference Voltage: {reference_voltage} V",
        ]
        if set_values:
            method = cls.methods[(method_byte >> 5) & 0b11]
            parts.insert(1, method)

        representation = ", ".join(parts)

        return representation


# -- Main ---------------------------------------------------------------------

if __name__ == "__main__":
    from doctest import testmod

    testmod()



"""Support for creating a connection to the ICOtronic system"""

# -- Imports ------------------------------------------------------------------

from __future__ import annotations

from asyncio import get_running_loop, to_thread
from sys import platform
from types import TracebackType

from can import Bus, BusABC, Notifier
from can.interfaces.pcan.pcan import PcanError

from icotronic.can.error import CANInitError
from icotronic.can.listener import Logger
from icotronic.can.node.spu import SPU
from icotronic.can.node.stu import STU
from icotronic.config import settings
from icotronic.test.misc import skip_hardware_tests_ci

pytestmark = skip_hardware_tests_ci()


# -- Classes ------------------------------------------------------------------


class Connection:
    """Basic class to initialize CAN communication

    To actually connect to the CAN bus you need to use the async context
    manager, provided by this class. If you want to manage the connection
    yourself, please just use ``__aenter__`` and ``__aexit__`` manually.

    Examples:

        Create a new connection (without connecting to the CAN bus)

        >>> connection = Connection()

    """

    def __init__(self) -> None:
        self.configuration = (
            settings.can.linux
            if platform == "linux"
            else (
                settings.can.mac
                if platform == "darwin"
                else settings.can.windows
            )
        )
        self.bus: BusABC | None = None
        self.notifier: Notifier | None = None

    async def __aenter__(self) -> STU:
        """Connect to the STU

        Returns:

            An object that can be used to communicate with the STU

        Raises:

            CANInitError: if the CAN initialization fails

        Examples:

            Import required library code

            >>> from asyncio import run

            Use a context manager to handle the cleanup process automatically

            >>> async def connect_can_context():
            ...     async with Connection() as stu:
            ...         pass
            >>> run(connect_can_context())

            Create and shutdown the connection explicitly

            >>> async def connect_can_manual():
            ...     connection = Connection()
            ...     connected = await connection.__aenter__()
            ...     await connection.__aexit__(None, None, None)
            >>> run(connect_can_manual())

        """

        def init():
            try:
                self.bus = Bus(  # pylint: disable=abstract-class-instantiated
                    channel=self.configuration.get("channel"),
                    interface=self.configuration.get("interface"),
                    bitrate=self.configuration.get("bitrate"),
                )  # type: ignore[abstract]
            except (PcanError, OSError) as error:
                raise CANInitError(
                    f"Unable to initialize CAN connection: {error}\n\n"
                    "Possible reasons:\n\n"
                    "• CAN adapter is not connected to the computer\n"
                    "• CAN adapter is in use by other program"
                ) from error

            self.bus.__enter__()

        await to_thread(init)

        assert isinstance(self.bus, BusABC)

        # We must set the event loop explicitly, otherwise the code will use
        # the synchronous API of python-can.
        self.notifier = Notifier(
            self.bus, listeners=[Logger()], loop=get_running_loop()
        )

        return STU(SPU(self.bus, self.notifier))

    async def __aexit__(
        self,
        exception_type: type[BaseException] | None,
        exception_value: BaseException | None,
        traceback: TracebackType | None,
    ) -> None:
        """Disconnect CAN connection and clean up resources

        Args:

            exception_type:
                The type of the exception in case of an exception

            exception_value:
                The value of the exception in case of an exception

            traceback:
                The traceback in case of an exception

        """

        notifier = self.notifier
        if notifier is not None:
            await to_thread(notifier.stop)

        bus = self.bus
        if bus is not None:
            await to_thread(bus.shutdown)


# -- Main ---------------------------------------------------------------------

if __name__ == "__main__":
    from doctest import testmod

    testmod()



"""Basic EEPROM support for ICOtronic nodes"""

# -- Imports ------------------------------------------------------------------

from struct import pack, unpack

from icotronic.can.protocol.message import Message
from icotronic.can.node.id import NodeId
from icotronic.can.node.spu import SPU

from icotronic.test.misc import skip_hardware_tests_ci
from icotronic.utility.data import convert_bytes_to_text

pytestmark = skip_hardware_tests_ci()


# -- Classes ------------------------------------------------------------------


class EEPROM:
    """Read and write EEPROM data of ICOtronic nodes

    Args:

        spu:
            A SPU object used to communicate with the ICOtronic system

        node:
            The node identifier of the node that contains the EEPROM

    """

    def __init__(self, spu: SPU, node: NodeId) -> None:

        self.spu = spu
        self.id = node

    async def read(self, address: int, offset: int, length: int) -> list[int]:
        """Read EEPROM data

        Args:

            address:
                The page number in the EEPROM

            offset:
                The offset to the base address in the specified page

            length:
                This value specifies how many bytes you want to read

        Returns:

            A list containing the EEPROM data at the specified location

        Examples:

            Import required library code

            >>> from asyncio import run
            >>> from icotronic.can.connection import Connection

            Read EEPROM data from STU 1

            >>> async def read_eeprom():
            ...     async with Connection() as stu:
            ...         return await stu.eeprom.read(
            ...             address=0, offset=1, length=8)
            >>> data = run(read_eeprom())
            >>> len(data)
            8
            >>> all((0 <= byte <= 255 for byte in data))
            True

        """

        read_data: list[int] = []
        reserved = [0] * 5
        data_start = 4  # Start index of data in response message

        node = self.id
        while length > 0:
            # Read at most 4 bytes of data at once
            read_length = 4 if length > 4 else length
            message = Message(
                block="EEPROM",
                block_command="Read",
                sender=self.spu.id,
                receiver=node,
                request=True,
                data=[address, offset, read_length, *reserved],
            )

            response = await self.spu.request(
                message, description=f"read EEPROM data from “{node}”"
            )

            data_end = data_start + read_length
            read_data.extend(response.data[data_start:data_end])
            length -= read_length
            offset += read_length

        return read_data

    async def read_float(self, address: int, offset: int) -> float:
        """Read EEPROM data in float format

        Args:

            address:
                The page number in the EEPROM

            offset:
                The offset to the base address in the specified page

        Returns:

            The float number at the specified location of the EEPROM

        Examples:

            Import required library code

            >>> from asyncio import run
            >>> from icotronic.can.connection import Connection
            >>> from icotronic.can.node.sth import STH

            Read slope of acceleration for x-axis of STH 1

            >>> async def read_slope():
            ...     async with Connection() as stu:
            ...         # We assume that at least one sensor node is available
            ...         async with stu.connect_sensor_node(0, STH) as sth:
            ...             return await sth.eeprom.read_float(
            ...                 address=8, offset=0)
            >>> slope = run(read_slope())
            >>> isinstance(slope, float)
            True

        """

        data = await self.read(address, offset, length=4)
        return unpack("<f", bytearray(data))[0]

    async def read_int(
        self,
        address: int,
        offset: int,
        length: int,
        signed: bool = False,
    ) -> int:
        """Read an integer value from the EEPROM

        Args:

            address:
                The page number in the EEPROM

            offset:
                The offset to the base address in the specified page

            length:
                This value specifies how long the number is in bytes

            signed:
                Specifies if `value` is a signed number (`True`) or an
                unsigned number (`False`)

        Returns:

            The number at the specified location of the EEPROM

        Examples:

            Import required library code

            >>> from asyncio import run
            >>> from icotronic.can.connection import Connection

            Read the operating time (in seconds) of STU 1

            >>> async def read_operating_time():
            ...     async with Connection() as stu:
            ...         return await stu.eeprom.read_int(address=5, offset=8,
            ...                                          length=4)
            >>> operating_time = run(read_operating_time())
            >>> operating_time >= 0
            True

        """

        return int.from_bytes(
            await self.read(address, offset, length),
            "little",
            signed=signed,
        )

    async def read_text(self, address: int, offset: int, length: int) -> str:
        """Read EEPROM data in ASCII format

        Please note, that this function will only return the characters up
        to the first null byte.

        Args:

            address:
                The page number in the EEPROM

            offset:
                The offset to the base address in the specified page

            length:
                This value specifies how many characters you want to read

        Returns:

            A string that contains the text at the specified location

        Examples:

            Import required library code

            >>> from asyncio import run
            >>> from icotronic.can.connection import Connection

            Read name of STU 1

            >>> async def read_name_eeprom():
            ...     async with Connection() as stu:
            ...         return await stu.eeprom.read_text(address=0, offset=1,
            ...                                           length=8)
            >>> name = run(read_name_eeprom())
            >>> 0 <= len(name) <= 8
            True
            >>> isinstance(name, str)
            True

        """

        data = await self.read(address, offset, length)
        return convert_bytes_to_text(data, until_null=True)

    async def write(
        self,
        address: int,
        offset: int,
        data: list[int],
        length: int | None = None,
    ) -> None:
        """Write EEPROM data at the specified address

        Args:

            address:
                The page number in the EEPROM

            offset:
                The offset to the base address in the specified page

            data:
                A list of byte value that should be stored at the specified
                EEPROM location

            length:
                This optional parameter specifies how many of the bytes in
                ``data`` should be stored in the EEPROM. If you specify a
                length that is greater, than the size of the data list, then
                the remainder of the EEPROM data will be filled with null
                bytes.

        Examples:

            Import required library code

            >>> from asyncio import run
            >>> from icotronic.can.connection import Connection

            Write data to and read (same) data from EEPROM of STU 1

            >>> async def write_and_read_eeprom(data):
            ...     async with Connection() as stu:
            ...         await stu.eeprom.write(address=10, offset=3, data=data)
            ...         return await stu.eeprom.read(address=10, offset=3,
            ...                                      length=len(data))
            >>> data = [1, 3, 3, 7]
            >>> read_data = run(write_and_read_eeprom(data))
            >>> data == read_data
            True

        """

        # Change data, if
        # - only a subset, or
        # - additional data
        # should be written to the EEPROM.
        if length is not None:
            # Cut off additional data bytes
            data = data[:length]
            # Fill up additional data bytes
            data.extend([0] * (length - len(data)))

        node = self.id

        while data:
            write_data = data[:4]  # Maximum of 4 bytes per message
            write_length = len(write_data)
            # Use zeroes to fill up missing data bytes
            write_data.extend([0] * (4 - write_length))

            reserved = [0] * 1
            message = Message(
                block="EEPROM",
                block_command="Write",
                sender=self.spu.id,
                receiver=node,
                request=True,
                data=[address, offset, write_length, *reserved, *write_data],
            )

            await self.spu.request(
                message, description=f"write EEPROM data in “{node}”"
            )

            data = data[4:]
            offset += write_length

    async def write_float(
        self,
        address: int,
        offset: int,
        value: float,
    ) -> None:
        """Write a float value at the specified EEPROM address

        Args:

            address:
                The page number in the EEPROM

            offset:
                The offset to the base address in the specified page

            value:
                The float value that should be stored at the specified location

        Examples:

            Import required library code

            >>> from asyncio import run
            >>> from icotronic.can.connection import Connection

            Write float value to and read (same) value from EEPROM of STU 1

            >>> async def write_and_read_float(value):
            ...     async with Connection() as stu:
            ...         await stu.eeprom.write_float(address=10, offset=0,
            ...                                      value=value)
            ...         return await stu.eeprom.read_float(
            ...             address=10, offset=0)
            >>> value = 42.5
            >>> read_value = run(write_and_read_float(value))
            >>> value == read_value
            True

        """

        data = list(pack("f", value))
        await self.write(address, offset, data)

    # pylint: disable=too-many-arguments, too-many-positional-arguments

    async def write_int(
        self,
        address: int,
        offset: int,
        value: int,
        length: int,
        signed: bool = False,
    ) -> None:
        """Write an integer number at the specified EEPROM address

        Args:

            address:
                The page number in the EEPROM

            offset:
                The offset to the base address in the specified page

            value:
                The number that should be stored at the specified location

            length:
                This value specifies how long the number is in bytes

            signed:
                Specifies if `value` is a signed number (`True`) or an
                unsigned number (`False`)

        Examples:

            Import required library code

            >>> from asyncio import run
            >>> from icotronic.can.connection import Connection

            Write int value to and read (same) int value from EEPROM of STU 1

            >>> async def write_and_read_int(value):
            ...     async with Connection() as stu:
            ...         await stu.eeprom.write_int(address=10, offset=0,
            ...             value=value, length=8, signed=True)
            ...         return await stu.eeprom.read_int(address=10, offset=0,
            ...                 length=8, signed=True)
            >>> value = -1337
            >>> read_value = run(write_and_read_int(value))
            >>> value == read_value
            True

        """

        data = list(value.to_bytes(length, byteorder="little", signed=signed))
        await self.write(address, offset, data)

    # pylint: enable=too-many-arguments, too-many-positional-arguments

    async def write_text(
        self,
        address: int,
        offset: int,
        text: str,
        length: int,
    ) -> None:
        """Write a string at the specified EEPROM address

        Args:

            address:
                The page number in the EEPROM

            offset:
                The offset to the base address in the specified page

            text:
                An ASCII string that should be written to the specified
                location

            length:
                This optional parameter specifies how many of the character in
                ``text`` should be stored in the EEPROM. If you specify a
                length that is greater than the size of the data list, then
                the remainder of the EEPROM data will be filled with null
                bytes.

        Examples:

            Import required library code

            >>> from asyncio import run
            >>> from icotronic.can.connection import Connection

            Write text to and read (same) text from EEPROM of STU 1

            >>> async def write_and_read_text(text):
            ...     async with Connection() as stu:
            ...         await stu.eeprom.write_text(address=10,
            ...                                     offset=11,
            ...                                     text=text,
            ...                                     length=len(text))
            ...         return await stu.eeprom.read_text(
            ...             address=10, offset=11, length=len(text))
            >>> run(write_and_read_text("something"))
            'something'

        """

        data = list(map(ord, list(text)))
        await self.write(address, offset, data, length)


# -- Main ---------------------------------------------------------------------

if __name__ == "__main__":
    from doctest import testmod

    testmod()



"""Support for the MyTooliT CAN protocol

See: https://mytoolit.github.io/Documentation/#mytoolit-communication-protocol

for more information
"""

# -- Exports ------------------------------------------------------------------

from icotronic.can.error import (
    ErrorResponseError,
    CANConnectionError,
    NoResponseError,
)
from icotronic.can.connection import Connection
from icotronic.can.streaming import (
    StreamingConfiguration,
    StreamingData,
    StreamingError,
    StreamingTimeoutError,
    StreamingBufferError,
)
from icotronic.can.sensor import SensorConfiguration
from icotronic.can.node.sensor import SensorNode
from icotronic.can.node.stu import STU
from icotronic.can.node.sth import STH



"""Support for the (29 bit) CAN identifiers of the MyTooliT protocol

See: https://mytoolit.github.io/Documentation/#identifier

for more information
"""

# -- Imports ------------------------------------------------------------------

from __future__ import annotations


from icotronic.can.protocol.command import Command
from icotronic.can.node.id import NodeId

# -- Class --------------------------------------------------------------------


class Identifier:
    """This class represents a CAN identifier of the MyTooliT protocol

    Usually you will either specify the identifier number directly, or
    provide command, sender and receiver. If, you decide to specify both
    the identifier value and one of the keyword arguments, then the
    keyword arguments will be used to overwrite specific parts of the
    identifier.

    Smaller parts (smaller bit width) will overwrite larger parts. For
    example, if you decide to both specify the command and the block
    (which is part of the command), then the block bits will be used to
    overwrite the block bits in the specified command.

    For more information, please take a look at the examples.

    Args:

        identifier:
            An extended CAN identifier (29 bit number)

        command:
            The whole command including group, number, error bit, and
            acknowledgement bit

        block:
            The block of the command (part of the command)

        block_command:
            The block command (part of the command)

        request:
            A boolean value that specifies if the identifier represents an
            request (or an acknowledgement)

        error:
            A boolean value that specifies if the identifier represents an
            error or not (part of the command)

        sender:
            The sender of the message

        receiver:
            The receiver of the message

    Examples:

        An empty identifier has value 0

        >>> Identifier().value
        0

        The class provided various methods to retrieve parts of the identifier

        >>> #                         V  block number   A E R send. R rec.
        >>> identifier = Identifier(0b0_000000_00000000_0_0_0_10000_0_00110,
        ...                         command=1337)
        >>> identifier.command()
        1337
        >>> identifier.receiver()
        6
        >>> identifier.sender()
        16

        >>> identifier = Identifier(command=512, sender='STH 1', receiver=2)
        >>> identifier.sender()
        1
        >>> identifier.receiver()
        2
        >>> identifier.command()
        512

        >>> #                         V  block number   A E R send. R rec.
        >>> identifier = Identifier(0b0_000000_00000000_1_0_0_10000_0_00110)
        >>> Identifier(identifier.value, request=False).is_acknowledgment()
        True

        >>> Identifier(receiver='SPU 1').receiver()
        15

        >>> identifier = Identifier(block=4, block_command=1, request=False,
        ...                         error=False, sender=31, receiver=1)
        >>> identifier.block_name()
        'Streaming'
        >>> identifier.block_command_name()
        'Temperature'

        >>> identifier = Identifier(block='EEPROM',
        ...                         block_command='Read')
        >>> identifier.block_name()
        'EEPROM'
        >>> identifier.block_command_name()
        'Read'

    """

    def __init__(  # pylint: disable=too-many-arguments
        self,
        *identifier: int,
        command: Command | None | int = None,
        block: None | str | int = None,
        block_command: None | str | int = None,
        error: bool | None = None,
        request: bool | None = None,
        sender: NodeId | None | str | int = None,
        receiver: NodeId | None | str | int = None,
    ) -> None:

        def set_part(start, width, number):
            """Store bit pattern number at bit start of the identifier"""

            identifier_ones = 0b11111_11111111_11111111_11111111
            mask = (1 << width) - 1

            # Set all bits for targeted part to 0
            self.value &= (mask << start) ^ identifier_ones
            # Make sure we use the correct number of bits for number
            number = number & mask
            # Set command bits to given value
            self.value |= number << start

        self.value = identifier[0] if identifier else 0

        if command is not None:
            # A command can be a number or a `Command` object
            command_as_number = (
                command if isinstance(command, int) else command.value
            )
            set_part(start=12, width=16, number=command_as_number)

        set_part(
            start=12,
            width=16,
            number=Command(
                (self.value >> 12) & 0xFFFF,
                block=block,
                block_command=block_command,
                request=request,
                error=error,
            ).value,
        )

        # Sender and receiver can be either an integer or a string like object
        if sender is not None:
            set_part(start=6, width=5, number=NodeId(sender).value)
        if receiver is not None:
            set_part(start=0, width=5, number=NodeId(receiver).value)

    def __eq__(self, other: object) -> bool:
        """Compare this identifier to another object

        Args:

            other:
                The other object this identifier should be compared to

        Returns:

            - ``True``, if the given object is an identifier and it has the
              same value as this identifier

            - ``False``, otherwise

        Examples:

            Compare some identifier objects for equality

            >>> identifier1 = Identifier(block='System', block_command='Reset')
            >>> identifier2 = Identifier(block='System', block_command='Reset',
            ...                          receiver='STH 1')

            >>> identifier1 == identifier2
            False

            >>> identifier1 == Identifier(identifier2.value, receiver=0)
            True

        """

        if isinstance(other, Identifier):
            return self.value == other.value

        return False

    def __repr__(self) -> str:
        """Return the string representation of the current identifier

        Returns:

            A string that describes the various attributes of the identifier

        Examples:

            Get the textual representation of some example identifiers

            >>> #            V  block number   A E R send. R rec.
            >>> Identifier(0b0_000000_00000000_0_0_0_00001_0_00010)
            [STH 1 → STH 2, Block: System, Command: Verboten, Acknowledge]

            >>> #                 V block  number   A E R send. R rec.
            >>> id = Identifier(0b0_000100_00000001_1_1_0_00010_0_00011)
            >>> id # doctest:+NORMALIZE_WHITESPACE
            [STH 2 → STH 3, Block: Streaming, Command: Temperature, Request,
             Error]

            >>> #            V block  number   A E R send. R rec.
            >>> Identifier(0b0_010101_10000001_1_1_0_00010_0_00011)
            [STH 2 → STH 3, Block: Unknown, Command: Unknown, Request, Error]

        """

        return (
            f"[{self.sender_name()} → {self.receiver_name()}, "
            f"{Command(self.command())}]"
        )

    def acknowledge(self, error: bool = False) -> Identifier:
        """Returns an acknowledgement identifier for this identifier

        In the acknowledgment identifier receiver and sender will be swapped
        and the request (acknowledge) bit will be set to 0 (acknowledge).

        Returns:

            An acknowledgment message for the current message

        Examples:

            Retrieve the acknowledgment identifier and check the result

            >>> identifier = Identifier(block='System', block_command='Reset',
            ...                         sender='SPU 1', receiver='STH 1')

            >>> acknowledgement_identifier = identifier.acknowledge()
            >>> acknowledgement_identifier.receiver_name()
            'SPU 1'
            >>> acknowledgement_identifier.is_error()
            False

            Retrieve the error identifier and check the result

            >>> error_identifier = identifier.acknowledge(error=True)
            >>> error_identifier.sender_name()
            'STH 1'
            >>> error_identifier.is_error()
            True

        """

        return Identifier(
            self.value,
            sender=self.receiver(),
            receiver=self.sender(),
            request=False,
            error=error,
        )

    def command(self) -> int:
        """Get the command part of the identifier

        Returns:

            The whole command including

            - group,
            - number,
            - acknowledge bit, and
            - error bit

            for the current identifier

        Examples:

            Retrieve the command number of an example identifier

            >>> #                V block  number   A E R send. R rec.
            >>> bin(Identifier(0b0_000011_00000000_0_1_0_00111_0_00010
            ...    ).command())
            '0b110000000001'

        """

        return (self.value >> 12) & 0xFFFF

    def command_number(self) -> int:
        """Get the block and block command part of the identifier

        Returns:

            The command without the error and acknowledge bits

        Examples:

            Retrieve the command number of an example identifier

            >>> #                V block  number   A E R send. R rec.
            >>> bin(Identifier(0b0_100011_11000001_0_1_0_00111_0_00010
            ...    ).command_number())
            '0b10001111000001'

        """

        return Command(self.command()).value >> 2

    def block(self) -> int:
        """Get the block

        Returns:

            The block (aka group) number for the current identifier

        Examples:

            Get the block number of a example identifier

            >>> #            V block  number   A E R send. R rec.
            >>> Identifier(0b0_000011_00000000_0_0_0_00111_0_00010).block()
            3

        """

        return Command(self.command()).block()

    def block_name(self) -> str:
        """Get the name of the command block

        Returns:

            A short textual description of the command block of the identifier

        Examples:

            >>> #            V  block  number  A E R send. R rec.
            >>> Identifier(0b0_000011_00000000_0_0_0_00111_0_00010
            ...           ).block_name()
            'Unknown'

            >>> #            V block  number   A E R send. R rec.
            >>> Identifier(0b0_000100_00000000_0_0_0_00111_0_00010
            ...           ).block_name()
            'Streaming'

        """

        return Command(self.command()).block_name()

    def block_command(self) -> int:
        """Get the block command

        Returns:

            The block number (part of the command number) of the identifier

        Examples:

            Retrieve the block command number of an example identifier

            >>> #            V  block number   A E R send. R rec.
            >>> Identifier(0b0_000011_00001000_0_0_0_00111_0_00010
            ...           ).block_command()
            8

        """

        return Command(self.command()).block_command()

    def block_command_name(self) -> str:
        """Get the name of the block command

        Returns:

            A short textual description of the command (in the current block)

        Examples:

            Retrieve the block command name of an example identifier

            >>> #            V block  number   A E R send. R rec.
            >>> Identifier(0b0_000000_00000000_0_0_0_00101_0_00010
            ...           ).block_command_name()
            'Verboten'

        """

        return Command(self.command()).block_command_name()

    def is_acknowledgment(self) -> bool:
        """Checks if the identifier represents an acknowledgment

        Returns:

            ``True`` if the identifier is for an acknowledgement, or ``False``
            otherwise

        Examples:

            Check if an example identifier represents an acknowledgment

            >>> #            V  block number   A E R send. R rec.
            >>> Identifier(0b0_000000_00000000_1_0_0_00101_0_00010
            ...           ).is_acknowledgment()
            False

        """

        return Command(self.command()).is_acknowledgment()

    def is_error(self) -> bool:
        """Checks if the identifier represents an error

        Returns:

            ``True`` if the identifier indicates an error message, or
            ``False`` otherwise

        Examples:

            Check if an example identifier represents an error

            >>> #            V block  number   A E R send. R rec.
            >>> Identifier(0b0_000000_00000000_1_1_0_00101_0_00010
            ...           ).is_error()
            True

        """

        return Command(self.command()).is_error()

    def set_acknowledgment(self, value: bool = True) -> Identifier:
        """Set the acknowledgment bit to the given value

        Args:

            value:
                A boolean that specifies if this identifier should represent an
                acknowledgment or not

        Examples:

            Set the acknowledgment bit and check the result

            >>> #                 V block  number   A E R send. R rec.
            >>> id = Identifier(0b0_000011_00001000_0_0_0_00111_0_00010)
            >>> id.is_acknowledgment()
            True
            >>> id.set_acknowledgment(False).is_acknowledgment()
            False

            >>> id.set_acknowledgment(True).is_acknowledgment()
            True

        Returns:

            The modified identifier object

        """

        command = Command(self.command()).set_acknowledgment(value)
        self.value = Identifier(self.value, command=command).value
        return self

    def set_error(self, error: bool = True) -> Identifier:
        """Set the error bit to the given value

        Args:

            error:
                A boolean that specifies if this identifier should indicate an
                error or not

        Examples:

            Set the error bit and check the result

            >>> #                 V block  number   A E R send. R rec.
            >>> id = Identifier(0b0_000011_00001000_0_1_0_00111_0_00010)
            >>> id.is_error()
            True

            >>> id.set_error(False).is_error()
            False

            >>> id.set_error().is_error()
            True

        """

        command = Command(self.command()).set_error(error)
        self.value = Identifier(self.value, command=command).value
        return self

    def sender(self) -> int:
        """Get the sender of the message

        Returns:

            A number that specifies the sender of the message

        Examples:

            Get the sender number of an example identifier

            >>> #            V  block   number A E R send. R rec.
            >>> Identifier(0b0_000000_00000000_0_0_0_00111_0_00010).sender()
            7


        """

        return self.value >> 6 & 0x1F

    def sender_name(self) -> str:
        """Get the name of the sender of a message

        Returns:

            A text that describes the sender

        Examples:

            Get the sender name of an example identifier

            >>> #            V block  number   A E R send. R rec.
            >>> Identifier(0b0_000000_00000000_0_0_0_00101_0_00010
            ...           ).sender_name()
            'STH 5'

        """

        return repr(NodeId(self.sender()))

    def receiver(self) -> int:
        """Get the receiver of the message

        Returns:

            A number that specifies the receiver of the message

        Examples:

            Get the receiver number of an example identifier

            >>> #            V  block number   A E R send. R rec.
            >>> Identifier(0b0_000000_00000000_0_0_0_00111_0_00010).receiver()
            2

        """

        return self.value & 0x1F

    def receiver_name(self) -> str:
        """Get the name of the receiver of a message

        Returns:

            A text that describes the receiver

        Examples:

            Get the receiver name of an example identifier

            >>> #            V block  number   A E R send. R rec.
            >>> Identifier(0b0_000000_00000000_0_0_0_00101_0_01110
            ...           ).receiver_name()
            'STH 14'

        """

        return repr(NodeId(self.receiver()))


# -- Main ---------------------------------------------------------------------

if __name__ == "__main__":
    from doctest import testmod

    testmod()



"""Support for adding forms (such as a checkbox) to PDF documents"""

# -- Imports ------------------------------------------------------------------

from __future__ import annotations

from reportlab.graphics.shapes import Drawing, Line, String
from reportlab.lib.colors import white
from reportlab.lib.units import cm
from reportlab.platypus import Flowable, KeepTogether, Paragraph, Table

from icotronic.report.style import get_style_sheet

# -- Classes ------------------------------------------------------------------

# pylint: disable=too-few-public-methods


class Signature:
    """This class represents a form for adding a signature"""

    def __init__(self) -> None:

        self.signature = Drawing(400, 30)

        style = get_style_sheet()["Normal"]

        indent = 180
        line_start = indent + 50
        line_end = line_start + 150
        self.signature.add(
            String(
                indent,
                0,
                "Signature:",
                fontName=style.fontName,
                fontSize=style.fontSize,
            )
        )
        self.signature.add(Line(line_start, 0, line_end, 0))

    def to_flowable(self) -> Drawing:
        """Convert the signature list into a Flowable

        Returns:

            A Flowable representing this signature

        """

        return self.signature


# pylint: enable=too-few-public-methods


class CheckBoxList:
    """This class represents a list of checkboxes

    The list starts with a title that should describe the purpose of the
    checkbox list, followed by the checkbox items. At the end a optional list
    of text fields will be added. These text fields can be used to add
    additional comments.

    Args:

        title:
            A title that describes the purpose of the checklist

        text_fields:
            The number of text fields for additional comments that should be
            added at the end of the checklist.

    """

    def __init__(self, title: str = "Checks", text_fields: int = 0) -> None:

        self.title = title
        self.checks: list[tuple[CheckBox, str]] = []
        self.styles = get_style_sheet()
        self.text_fields = text_fields

    def add_checkbox_item(self, text: str, tooltip: str | None = None) -> None:
        """Add a checkbox item to the checkbox list

        Args:

            text:
                The text that should be added after the checkbox item

            tooltip:
                The tooltip for the checkbox; If you do not specify a tooltip,
                then ``text`` will also be used for the tooltip.

        """

        tooltip = text if tooltip is None else tooltip
        self.checks.append((CheckBox(tooltip), text))

    def to_flowable(self) -> KeepTogether:
        """Convert the checkbox list into a Flowable

        Returns:

            A ``Flowable`` representing this checkbox list

        """

        title = Paragraph(self.title, style=self.styles["Heading3"])

        # Somehow the text columns of a table will contain a lot of
        # trailing whitespace, if some (other) cells contain non-textual
        # data. We work around that by specifying the size of the first
        # column manually.
        checks = Table(self.checks, colWidths=[0.5 * cm, None])

        text_fields = [
            TextField("Additional comments") for _ in range(self.text_fields)
        ]

        return KeepTogether([title, checks, *text_fields])


class CheckBox(Flowable):
    """A flowable checkbox

    Args:

        tooltip:
            The text displayed in the tooltip of the checkbox

    Examples:

        Create some example checkboxes

        >>> CheckBox(tooltip="The tooltip of the box")
        ☑️ Tooltip: The tooltip of the box

        >>> CheckBox()
        ☑️

    """

    def __init__(self, tooltip: str | None = None) -> None:

        super().__init__()

        self.tooltip = tooltip
        self.boxsize = 10

        self.width = self.height = self.boxsize

    def __repr__(self) -> str:
        """The string representation of the checkbox

        Returns:

            A string containing information about the checkbox

        """

        text = f" Tooltip: {self.tooltip}" if self.tooltip is not None else ""
        return f"☑️{text}"

    def draw(self) -> None:
        """Draw the checkbox on the canvas"""

        self.canv.saveState()

        form = self.canv.acroForm
        form.checkbox(
            checked=False,
            buttonStyle="check",
            fillColor=white,
            tooltip=self.tooltip,
            relative=True,
            size=self.boxsize,
        )

        self.canv.restoreState()


class TextField(Flowable):
    """A flowable text box

    Args:

        tooltip:
            The text displayed in the tooltip of the text

    Examples:

        Create an example text field

        >>> TextField(tooltip="The tooltip of the text field")
        📝 Tooltip: The tooltip of the text field

    """

    def __init__(self, tooltip: str | None = None) -> None:

        super().__init__()

        self.tooltip = tooltip

        self.indent = 6  # Indent slightly to match indentation of checkboxes
        self.width = 350
        self.height = 18
        self.styles = get_style_sheet()

    def __repr__(self) -> str:
        """The string representation of the checkbox

        Returns:

            A string containing information about the checkbox

        """

        text = f" Tooltip: {self.tooltip}" if self.tooltip is not None else ""
        return f"📝{text}"

    def draw(self) -> None:
        """Draw the text box on the canvas"""

        self.canv.saveState()

        form = self.canv.acroForm
        style = self.styles["Normal"]
        form.textfield(
            x=self.indent,
            fontName=style.fontName,
            fontSize=style.fontSize,
            fillColor=white,
            tooltip=self.tooltip,
            relative=True,
            borderWidth=0.5,
            width=self.width,
            height=self.height,
        )

        self.canv.restoreState()


# -- Main ---------------------------------------------------------------------

if __name__ == "__main__":
    from doctest import testmod

    testmod()



"""Measurement support code"""

# -- Imports ------------------------------------------------------------------

from __future__ import annotations

from typing import Callable, Iterator, NamedTuple

from icotronic.can.streaming.config import StreamingConfiguration
from icotronic.can.streaming.data import StreamingData
from icotronic.can.dataloss import calculate_dataloss_stats

# -- Classes ------------------------------------------------------------------


class DataPoint(NamedTuple):
    """Streaming data point"""

    counter: int
    """Message counter"""
    timestamp: float
    """Timestamp of data"""
    value: float
    """Data value"""

    def __repr__(self):
        """Get textual representation of data

        Returns:

            A string containing the attributes of this datapoint

        Examples:

            Get the textual representation of a data point

            >>> DataPoint(counter=1, timestamp=2, value=3)
            3@2 #1

        """

        return f"{self.value}@{self.timestamp} #{self.counter}"


class ChannelData:
    """Store measurement data for a single channel"""

    def __init__(self) -> None:
        self.data: list[DataPoint] = []

    def __repr__(self) -> str:
        """Get the string representation of the data

        Examples:

            Get string representation of some example data

            >>> t1 = 1756124450.256398
            >>> t2 = 1756124450.2564
            >>> data = ChannelData()
            >>> data.append(DataPoint(counter=1, timestamp=t1, value=4))
            >>> data.append(DataPoint(counter=1, timestamp=t1, value=5))
            >>> data.append(DataPoint(counter=1, timestamp=t1, value=6))
            >>> data.append(DataPoint(counter=2, timestamp=t2, value=7))
            >>> data.append(DataPoint(counter=2, timestamp=t2, value=8))
            >>> data.append(DataPoint(counter=2, timestamp=t2, value=9))

            >>> data # doctest:+NORMALIZE_WHITESPACE
            4@1756124450.256398 #1
            5@1756124450.256398 #1
            6@1756124450.256398 #1
            7@1756124450.2564 #2
            8@1756124450.2564 #2
            9@1756124450.2564 #2

        """

        return "\n".join([repr(datapoint) for datapoint in self.data])

    def __iter__(self) -> Iterator:
        """Iterate over the channel data

        Returns:

            An iterator over the data points of the channel data

        Examples:

            Print the timestamps of some channel data

            >>> data = ChannelData()
            >>> t1 = 1756124450.1234
            >>> t2 = 1756124450.1235

            >>> data.append(DataPoint(counter=1, timestamp=t1, value=4))
            >>> data.append(DataPoint(counter=1, timestamp=t2, value=5))

            >>> for point in data:
            ...     print(point.timestamp)
            1756124450.1234
            1756124450.1235

        """

        return iter(self.data)

    def append(self, data: DataPoint) -> None:
        """Append a value to the channel data

        Args:

            data:

                The data point that should be added to the channel data

        Examples:

            Add some data points to a channel data object

            >>> data = ChannelData()

            >>> t1 = 1756124450.256398
            >>> t2 = t1 + 0.000003
            >>> data.append(DataPoint(counter=255, timestamp=t1, value=10))
            >>> data.append(DataPoint(counter=1, timestamp=t2, value=20))
            >>> data
            10@1756124450.256398 #255
            20@1756124450.256401 #1

        """

        self.data.append(data)


# pylint: disable=too-few-public-methods


class Conversion:
    """Conversion functions for measurement data

    Args:

        first:

            The conversion function for the first channel

        second:

            The conversion function for the second channel

        third:

            The conversion function for the third channel

    """

    def __init__(
        self,
        first: Callable[[float], float] | None = None,
        second: Callable[[float], float] | None = None,
        third: Callable[[float], float] | None = None,
    ) -> None:
        self.first = first
        self.second = second
        self.third = third


# pylint: enable=too-few-public-methods


class MeasurementData:
    """Measurement data

    Args:

        configuration:

            The streaming configuration that was used to collect the
            measurement data

    """

    def __init__(self, configuration: StreamingConfiguration) -> None:
        self.configuration = configuration
        self.streaming_data_list: list[StreamingData] = []

    def __repr__(self) -> str:
        """Get the textual representation of the measurement data

        Returns:

            The textual representation of the measurement data

        Examples:

            >>> config = StreamingConfiguration(first=True, second=True,
            ...                                 third=False)
            >>> data = MeasurementData(config)
            >>> s1 = StreamingData(values=[1, 2], counter=255,
            ...                    timestamp=1756125747.528234)
            >>> s2 = StreamingData(values=[3, 4], counter=0,
            ...                    timestamp=1756125747.528237)
            >>> data.append(s1)

            >>> data
            Channel 1 enabled, Channel 2 enabled, Channel 3 disabled
            [1, 2]@1756125747.528234 #255

            >>> data.append(s2)
            >>> data
            Channel 1 enabled, Channel 2 enabled, Channel 3 disabled
            [1, 2]@1756125747.528234 #255
            [3, 4]@1756125747.528237 #0

        """

        return (
            f"{self.configuration}"
            + ("\n" if self.streaming_data_list else "")
            + "\n".join([
                str(streaming_data)
                for streaming_data in self.streaming_data_list
            ])
        )

    def __iter__(self) -> Iterator:
        """Iterate over the measurement data

        Returns:

            An iterator over the streaming dat objects contained in the
            measurement data

        Examples:

            Iterate over some example measurement data

            >>> config = StreamingConfiguration(first=False, second=True,
            ...                                 third=False)
            >>> data = MeasurementData(config)
            >>> s1 = StreamingData(values=[123, 456], counter=155,
            ...                    timestamp=1758029094.7407959)
            >>> s2 = StreamingData(values=[78, 9], counter=156,
            ...                    timestamp=1758029094.7407968)
            >>> data.append(s1)
            >>> data.append(s2)

            >>> for stream_data in data:
            ...     print(stream_data)
            [123, 456]@1758029094.7407959 #155
            [78, 9]@1758029094.7407968 #156

        """

        return iter(self.streaming_data_list)

    def __len__(self) -> int:
        """Get the length of the measurement data

        The length is defined as the number of streaming data items contained
        in the measurement data.

        Returns:

            The number of streaming data elements in the measurement data

        Examples:

            >>> config = StreamingConfiguration(first=True, second=True,
            ...                                 third=False)
            >>> data = MeasurementData(config)
            >>> len(data)
            0

            >>> s1 = StreamingData(values=[1, 2], counter=255,
            ...                    timestamp=1756125747.528234)
            >>> data.append(s1)
            >>> len(data)
            1

            >>> s2 = StreamingData(values=[3, 4], counter=0,
            ...                    timestamp=1756125747.528237)
            >>> data.append(s2)
            >>> len(data)
            2

        """

        return len(self.streaming_data_list)

    def first(self) -> ChannelData:
        """Get all data of the first measurement channel

        Returns:

            Data values for the first measurement channel

        Examples:

            Get first channel data of measurement with two enabled channels

            >>> config = StreamingConfiguration(first=True, second=True,
            ...                                 third=False)
            >>> data = MeasurementData(config)
            >>> s1 = StreamingData(values=[1, 2], counter=255,
            ...                    timestamp=1756125747.528234)
            >>> s2 = StreamingData(values=[3, 4], counter=0,
            ...                    timestamp=1756125747.528237)
            >>> data.append(s1)
            >>> data.append(s2)
            >>> data.first() # doctest:+NORMALIZE_WHITESPACE
            1@1756125747.528234 #255
            3@1756125747.528237 #0
            >>> data.second() # doctest:+NORMALIZE_WHITESPACE
            2@1756125747.528234 #255
            4@1756125747.528237 #0
            >>> data.third()
            <BLANKLINE>

            Get first channel data of measurement with one enabled channel

            >>> config = StreamingConfiguration(first=True, second=False,
            ...                                 third=False)
            >>> data = MeasurementData(config)
            >>> s1 = StreamingData(values=[1, 2, 3], counter=10,
            ...                    timestamp=1756126628.820695)
            >>> s2 = StreamingData(values=[4, 5, 6], counter=20,
            ...                    timestamp=1756126628.8207)
            >>> data.append(s1)
            >>> data.append(s2)
            >>> data.first() # doctest:+NORMALIZE_WHITESPACE
            1@1756126628.820695 #10
            2@1756126628.820695 #10
            3@1756126628.820695 #10
            4@1756126628.8207 #20
            5@1756126628.8207 #20
            6@1756126628.8207 #20
            >>> data.second()
            <BLANKLINE>
            >>> data.third()
            <BLANKLINE>

        """

        channel_data = ChannelData()
        configuration = self.configuration

        if configuration.first:
            if not configuration.second and not configuration.third:
                # Three values
                for streaming_data in self.streaming_data_list:
                    counter = streaming_data.counter
                    timestamp = streaming_data.timestamp
                    for value in streaming_data.values:
                        channel_data.append(
                            DataPoint(
                                counter=counter,
                                timestamp=timestamp,
                                value=value,
                            )
                        )
            else:
                # One value
                for streaming_data in self.streaming_data_list:
                    channel_data.append(
                        DataPoint(
                            counter=streaming_data.counter,
                            timestamp=streaming_data.timestamp,
                            value=streaming_data.values[0],
                        )
                    )

        return channel_data

    def second(self) -> ChannelData:
        """Get all data of the second measurement channel

        Returns:

            Data values for the second measurement channel

        Examples:

            Get second channel data of measurement with two enabled channels

            >>> config = StreamingConfiguration(first=False, second=True,
            ...                                 third=True)
            >>> data = MeasurementData(config)
            >>> s1 = StreamingData(values=[1, 2], counter=255,
            ...                    timestamp=1756125747.528234)
            >>> s2 = StreamingData(values=[3, 4], counter=0,
            ...                    timestamp=1756125747.528237)
            >>> data.append(s1)
            >>> data.append(s2)
            >>> data.first()
            <BLANKLINE>
            >>> data.second() # doctest:+NORMALIZE_WHITESPACE
            1@1756125747.528234 #255
            3@1756125747.528237 #0
            >>> data.third() # doctest:+NORMALIZE_WHITESPACE
            2@1756125747.528234 #255
            4@1756125747.528237 #0

            Get second channel data of measurement with one enabled channel

            >>> config = StreamingConfiguration(first=False, second=True,
            ...                                 third=False)
            >>> data = MeasurementData(config)
            >>> s1 = StreamingData(values=[1, 2, 3], counter=10,
            ...                    timestamp=1756126628.820695)
            >>> s2 = StreamingData(values=[4, 5, 6], counter=20,
            ...                    timestamp=1756126628.8207)
            >>> data.append(s1)
            >>> data.append(s2)
            >>> data.first()
            <BLANKLINE>
            >>> data.second() # doctest:+NORMALIZE_WHITESPACE
            1@1756126628.820695 #10
            2@1756126628.820695 #10
            3@1756126628.820695 #10
            4@1756126628.8207 #20
            5@1756126628.8207 #20
            6@1756126628.8207 #20
            >>> data.third()
            <BLANKLINE>

        """

        channel_data = ChannelData()
        configuration = self.configuration

        if configuration.second:
            if not configuration.first and not configuration.third:
                # Three values
                for streaming_data in self.streaming_data_list:
                    counter = streaming_data.counter
                    timestamp = streaming_data.timestamp
                    for value in streaming_data.values:
                        channel_data.append(
                            DataPoint(
                                counter=counter,
                                timestamp=timestamp,
                                value=value,
                            )
                        )
            else:
                # One value
                for streaming_data in self.streaming_data_list:
                    channel_data.append(
                        DataPoint(
                            counter=streaming_data.counter,
                            timestamp=streaming_data.timestamp,
                            value=(
                                streaming_data.values[0]
                                if not configuration.first
                                else streaming_data.values[1]
                            ),
                        )
                    )

        return channel_data

    def third(self) -> ChannelData:
        """Get all data of the third measurement channel

        Returns:

            Data values for the third measurement channel

        Examples:

            Get third channel data of measurement with two enabled channels

            >>> config = StreamingConfiguration(first=True, second=False,
            ...                                 third=True)
            >>> data = MeasurementData(config)
            >>> s1 = StreamingData(values=[1, 2], counter=255,
            ...                    timestamp=1756125747.528234)
            >>> s2 = StreamingData(values=[3, 4], counter=0,
            ...                    timestamp=1756125747.528237)
            >>> data.append(s1)
            >>> data.append(s2)
            >>> data.first() # doctest:+NORMALIZE_WHITESPACE
            1@1756125747.528234 #255
            3@1756125747.528237 #0
            >>> data.second()
            <BLANKLINE>
            >>> data.third() # doctest:+NORMALIZE_WHITESPACE
            2@1756125747.528234 #255
            4@1756125747.528237 #0

            Get third channel data of measurement with one enabled channel

            >>> config = StreamingConfiguration(first=False, second=False,
            ...                                 third=True)
            >>> data = MeasurementData(config)
            >>> s1 = StreamingData(values=[1, 2, 3], counter=10,
            ...                    timestamp=1756126628.820695)
            >>> s2 = StreamingData(values=[4, 5, 6], counter=20,
            ...                    timestamp=1756126628.8207)
            >>> data.append(s1)
            >>> data.append(s2)
            >>> data.first()
            <BLANKLINE>
            >>> data.second()
            <BLANKLINE>
            >>> data.third() # doctest:+NORMALIZE_WHITESPACE
            1@1756126628.820695 #10
            2@1756126628.820695 #10
            3@1756126628.820695 #10
            4@1756126628.8207 #20
            5@1756126628.8207 #20
            6@1756126628.8207 #20

        """

        channel_data = ChannelData()
        configuration = self.configuration

        if configuration.third:
            if not configuration.first and not configuration.second:
                # Three values
                for streaming_data in self.streaming_data_list:
                    counter = streaming_data.counter
                    timestamp = streaming_data.timestamp
                    for value in streaming_data.values:
                        channel_data.append(
                            DataPoint(
                                counter=counter,
                                timestamp=timestamp,
                                value=value,
                            )
                        )
            else:
                # One value
                for streaming_data in self.streaming_data_list:
                    channel_data.append(
                        DataPoint(
                            counter=streaming_data.counter,
                            timestamp=streaming_data.timestamp,
                            value=streaming_data.values[-1],
                        )
                    )

        return channel_data

    def dataloss(self) -> float:
        """Get measurement dataloss based on message counters

        Returns:

            The overall amount of dataloss as number between 0 (no data loss)
            and 1 (all data lost).

        """

        return calculate_dataloss_stats((
            streaming_data.counter
            for streaming_data in self.streaming_data_list
        )).dataloss()

    def append(self, data: StreamingData) -> None:
        """Append some streaming data to the measurement

        Args:

            data:

                The streaming data that should be added to the measurement

        Examples:

            Append some streaming data to a measurement

            >>> config = StreamingConfiguration(first=True, second=True,
            ...                                 third=True)
            >>> data = MeasurementData(config)
            >>> data
            Channel 1 enabled, Channel 2 enabled, Channel 3 enabled

            >>> s1 = StreamingData(values=[4, 5, 3], counter=15,
            ...                    timestamp=1756197008.776551)
            >>> data.append(s1)
            >>> data
            Channel 1 enabled, Channel 2 enabled, Channel 3 enabled
            [4, 5, 3]@1756197008.776551 #15

        """

        self.streaming_data_list.append(data)

    def extend(self, data: MeasurementData) -> None:
        """Extend this measurement data with some other measurement data

        Args:

            data:

                The measurement data that should be added to this measurement

        Examples:

            Extend measurement data with other measurement data

            >>> config = StreamingConfiguration(first=True, second=False,
            ...                                 third=True)
            >>> data1 = MeasurementData(config)
            >>> s1 = StreamingData(values=[1, 2], counter=255,
            ...                    timestamp=1756125747.528234)
            >>> s2 = StreamingData(values=[3, 4], counter=0,
            ...                    timestamp=1756125747.528237)
            >>> data1.append(s1)
            >>> data1.append(s2)
            >>> data1
            Channel 1 enabled, Channel 2 disabled, Channel 3 enabled
            [1, 2]@1756125747.528234 #255
            [3, 4]@1756125747.528237 #0

            >>> data2 = MeasurementData(config)
            >>> s3 = StreamingData(values=[10, 20], counter=1,
            ...                    timestamp=1756125747.678912)
            >>> data2.append(s3)
            >>> data2
            Channel 1 enabled, Channel 2 disabled, Channel 3 enabled
            [10, 20]@1756125747.678912 #1

            >>> data1.extend(data2)
            >>> data1
            Channel 1 enabled, Channel 2 disabled, Channel 3 enabled
            [1, 2]@1756125747.528234 #255
            [3, 4]@1756125747.528237 #0
            [10, 20]@1756125747.678912 #1

        """

        if self.configuration != data.configuration:
            raise ValueError(
                f"Trying to merge measurement data {self.configuration} with "
                "different streaming configuration: {data.configuration}"
            )

        self.streaming_data_list.extend(data.streaming_data_list)

    def apply(self, conversion: Conversion) -> MeasurementData:
        """Apply functions to the values stored in the measurement

        Args:

            conversion:

                The conversion functions that will be applied to the
                measurement

        Returns:

            The measurement data itself, after the conversion was applied

        Examples:

            Apply functions to some measurement data with one active channel


            >>> config = StreamingConfiguration(first=False, second=True,
            ...                                 third=False)
            >>> data = MeasurementData(config)
            >>> s1 = StreamingData(values=[1, 20, 81], counter=22,
            ...                    timestamp=1756125747.528234)
            >>> s2 = StreamingData(values=[50, 1, 29], counter=25,
            ...                    timestamp=1756125747.528254)
            >>> data.append(s1)
            >>> data.append(s2)

            >>> plus_two = (lambda value: value + 2)
            >>> conversion = Conversion(second=plus_two)

            >>> data
            Channel 1 disabled, Channel 2 enabled, Channel 3 disabled
            [1, 20, 81]@1756125747.528234 #22
            [50, 1, 29]@1756125747.528254 #25
            >>> data.apply(conversion)
            Channel 1 disabled, Channel 2 enabled, Channel 3 disabled
            [3, 22, 83]@1756125747.528234 #22
            [52, 3, 31]@1756125747.528254 #25

            Apply functions to some measurement data with two active channels

            >>> config = StreamingConfiguration(first=True, second=False,
            ...                                 third=True)
            >>> data = MeasurementData(config)
            >>> s1 = StreamingData(values=[1, 2], counter=255,
            ...                    timestamp=1756125747.528234)
            >>> s2 = StreamingData(values=[3, 4], counter=0,
            ...                    timestamp=1756125747.528237)
            >>> data.append(s1)
            >>> data.append(s2)

            >>> conversion = Conversion(third=plus_two)

            >>> data
            Channel 1 enabled, Channel 2 disabled, Channel 3 enabled
            [1, 2]@1756125747.528234 #255
            [3, 4]@1756125747.528237 #0

            >>> data.apply(conversion)
            Channel 1 enabled, Channel 2 disabled, Channel 3 enabled
            [1, 4]@1756125747.528234 #255
            [3, 6]@1756125747.528237 #0

            Apply functions to some measurement data with three active channels

            >>> config = StreamingConfiguration(first=True, second=True,
            ...                                 third=True)
            >>> data = MeasurementData(config)
            >>> s1 = StreamingData(values=[4, 5, 3], counter=15,
            ...                    timestamp=1756197008.776551)
            >>> s2 = StreamingData(values=[8, 10, 6], counter=16,
            ...                    timestamp=1756197008.776559)
            >>> data.append(s1)
            >>> data.append(s2)

            >>> double = (lambda value: value * 2)
            >>> conversion = Conversion(first=double, third=plus_two)

            >>> data
            Channel 1 enabled, Channel 2 enabled, Channel 3 enabled
            [4, 5, 3]@1756197008.776551 #15
            [8, 10, 6]@1756197008.776559 #16

            >>> data.apply(conversion)
            Channel 1 enabled, Channel 2 enabled, Channel 3 enabled
            [8, 5, 5]@1756197008.776551 #15
            [16, 10, 8]@1756197008.776559 #16

        """

        enabled_channels = self.configuration.enabled_channels()
        config = self.configuration

        functions: list[Callable[[float], float] | None] = []

        if enabled_channels == 1:
            # Apply single function to all values
            function = (
                conversion.first
                if config.first
                else (conversion.second if config.second else conversion.third)
            )
            if function is not None:
                functions = [function, function, function]

        elif enabled_channels == 2:
            # Apply single function to single value
            if config.first:
                functions = [
                    conversion.first,
                    conversion.second if config.second else conversion.third,
                ]
            else:
                functions = [conversion.second, conversion.third]

        elif enabled_channels == 3:
            functions = [conversion.first, conversion.second, conversion.third]

        for streaming_data in self.streaming_data_list:
            values = streaming_data.values
            for (index, value), function in zip(enumerate(values), functions):
                if function is not None:
                    values[index] = function(value)

        return self


# -- Main ---------------------------------------------------------------------

if __name__ == "__main__":
    from doctest import testmod

    testmod()
